from gwtc.gwtc_gracedb import GWTCGraceDB
from .gwtc import smap_from_gracedb_query
import logging
from optparse import OptionParser
import json

# setup verbose logs
logging.basicConfig(level=logging.INFO)


def parse_command_line():
    parser = OptionParser(usage="usage: %prog [options] query")
    parser.add_option(
        "--proxy-path", type="string", help="Set the path to find the proxy file"
    )
    parser.add_option(
        "--service-url",
        type="string",
        default="https://gracedb-test.ligo.org/api/",
        help="Set the gracedb service url. Default: https://gracedb-test.ligo.org/api/",
    )
    parser.add_option(
        "--number", type="str", default="4", help='Set the catalog number: default "4"'
    )
    options, query = parser.parse_args()

    return options, query


def main():
    options, query = parse_command_line()

    client = GWTCGraceDB(service_url=options.service_url, cred=options.proxy_path)
    query = query[0].replace("\\", " ")
    client = GWTCGraceDB(service_url=options.service_url, cred=options.proxy_path)
    query = query[0].replace("\\", " ")

    smap = smap_from_gracedb_query(client, query)

    resp = client.gwtc_create(smap, number=options.number)
    logging.info(f"Created GWTC:\n{json.dumps(resp.json(), indent=4)}")



import gwtc

# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "gwtc-tools"
copyright = "2023, Chad Hanna and Others"
author = "Chad Hanna and Others"

fullversion = gwtc.__version__
version = gwtc.__version__


# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "numpydoc",
    "sphinx_tabs.tabs",
    "sphinx.ext.mathjax",
    "sphinx.ext.autosummary",
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.viewcode",
]

templates_path = ["_templates"]
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_static_path = ["_static"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store", "requirements.txt"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

html_theme = "sphinx_rtd_theme"

html_theme_options = {
    "canonical_url": "",
    "logo_only": False,
    "display_version": True,
    "prev_next_buttons_location": "bottom",
    "style_external_links": False,
    # Toc options
    "collapse_navigation": True,
    "sticky_navigation": True,
    "navigation_depth": 4,
    "includehidden": True,
    "titles_only": False,
}

numpydoc_show_class_members = False

# Multiversion options
# Whitelist pattern for tags (set to None to ignore all tags)
smv_tag_whitelist = r"^(1.*|0.3.12)$"

# Whitelist pattern for branches (set to None to ignore all branches)
smv_branch_whitelist = r"^master$"

# Whitelist pattern for remotes (set to None to use local branches only)
smv_remote_whitelist = None

# Format for versioned output directories inside the build directory
smv_outputdir_format = "{ref.name}"

# Determines whether remote or local git branches/tags are preferred if their output dirs conflict
smv_prefer_remote_refs = False



import os
import shutil
import yaml

from typing import Optional, Union, List

from cbcflow.core.database import LocalLibraryDatabase
from cbcflow.core.utils import get_uid_index_for_object_array


class Aggregator:
    def __init__(self, path_to_recipe: Optional[str] = None) -> None:
        """Load configuration information from recipe.yml

        Parameters
        ==========
        path_to_recipe : Optional[str]
            The path to the recipe.yml file; if not passed assumes it can be found in the current working directory
        """
        if path_to_recipe is None:
            path_to_recipe = "recipe.yml"
        with open(path_to_recipe, "r") as f:
            recipe_contents = yaml.safe_load(f)
        self.output_directory = recipe_contents["general"]["output_directory"]
        self.cbcflow_library = recipe_contents["general"]["cbcflow_library"]
        self.cbcflow_library.load_library_metadata_dict()
        self.guaranteed_far_threshold = recipe_contents["parameter_estimation"][
            "guaranteed_far_threshold"
        ]
        self.best_effort_superevents = recipe_contents["parameter_estimation"][
            "best_effort_superevents"
        ]
        self._get_parameter_estimation_catalog_event_list()

    def _get_parameter_estimation_catalog_event_list(self):
        """Gets the events in the parameter estimation catalog, from those that past the guaranteed the far threshold
        plus those set as best effort events
        """
        self.parameter_estimation_catalog_event_list = list()
        for superevent, metadata in self.cbcflow_library.metadata_dict.items():
            # Read this off of the first event for now
            # In the future add field to schema to save value from the gwtc_get() table?
            superevent_far = float(metadata["GraceDB"]["Events"][0]["FAR"])
            if superevent_far <= self.guaranteed_far_threshold:
                self.parameter_estimation_catalog_event_list.append(superevent)
        self.parameter_estimation_catalog_event_list += self.best_effort_superevents

    @property
    def cbcflow_library(self) -> LocalLibraryDatabase:
        """The cbcflow library used to build the catalog product"""
        return self._cbcflow_library

    @cbcflow_library.setter
    def cbcflow_library(self, library: Union[str, LocalLibraryDatabase]):
        if isinstance(library, LocalLibraryDatabase):
            self._cbcflow_library = library
        elif isinstance(library, str):
            self._cbcflow_library = LocalLibraryDatabase(library)

    @property
    def best_effort_superevents(self) -> List[str]:
        return self._best_effort_superevents

    @best_effort_superevents.setter
    def best_effort_superevents(self, config_value: Union[None, List]) -> None:
        if config_value is None:
            self._best_effort_superevents = []
        else:
            self.best_effort_superevents = config_value

    @property
    def guaranteed_far_threshold(self) -> float:
        return self._guaranteed_far_threshold

    @guaranteed_far_threshold.setter
    def guaranteed_far_threshold(self, threshold) -> None:
        self._guaranteed_far_threshold = float(threshold)

    def fetch_illustrative_parameter_estimation_results(self):
        """Copies pesummary result .hdf5's for the illustrative result from each event in the pe catalog list"""
        for superevent in self.parameter_estimation_catalog_event_list:
            illustrative_result_id = self.cbcflow_library.metadata_dict[superevent][
                "ParameterEstimation"
            ]["IllustrativeResult"]
            # Should be pesummary results in the future! But that's not accessible presently with ROTA results (should just be an addition to the monitor?)
            # So save bilby result file instead for now
            results = self.cbcflow_library.metadata_dict[superevent][
                "ParameterEstimation"
            ]["Results"]
            try:
                illustrative_result_index = get_uid_index_for_object_array(
                    illustrative_result_id, results
                )
            except ValueError:
                print(
                    f"For superevent {superevent} IllustrativeResult mixes upper and lower case, but the UID is uppercase"
                )
                # temporary fix for some IllustrativeResult being e.g. 'Exp2' when the UID is 'EXP2'
                illustrative_result_index = get_uid_index_for_object_array(
                    illustrative_result_id.upper(), results
                )
            # We're assuming that both we and the result are on CIT
            try:
                illustrative_result_path = results[illustrative_result_index][
                    "ResultFile"
                ]["Path"].replace("CIT:", "")
            except KeyError:
                print(
                    f"For superevent {superevent} couldn't find preferred ResultFile, defaulting to online result"
                )
                # Some ROTA uploads are broken, in the meantime do this
                illustrative_result_index = get_uid_index_for_object_array(
                    "online", results
                )
                illustrative_result_path = results[illustrative_result_index][
                    "ResultFile"
                ]["Path"].replace("CIT:", "")

            shutil.copy(
                illustrative_result_path,
                os.path.join(self.output_directory, f"{superevent}-result.hdf5"),
            )


def main():
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("recipe", type=str, help="The recipe for data aggregation")
    args = parser.parse_args()

    aggregator = Aggregator(path_to_recipe=args.recipe)
    aggregator.fetch_illustrative_parameter_estimation_results()



#!/usr/bin/env python3
# import fake_gracedb
from gwtc.gwtc_gracedb import GWTCGraceDB

# import sqlite3
import logging
import json
import requests

from typing import Dict, Tuple

# setup verbose logs
logging.basicConfig(level=logging.INFO)

#
# Utility functions to define catalogs directly from useful gracedb queries. These are mostly for demos
#


def update_far_and_pastro_in_smap(gdb: GWTCGraceDB, smap: Dict) -> Dict:
    """
    Parameters
    ==========
    gdb : `gwtc.gwtc_gracedb.GWTCGraceDB`
        A rest API GraceDB instance modified for GWTC support
    smap : dict
        e.g.,

        "S230824d": {
            "pipelines": {
                "gstlal": "G651775"
            },
            "far": null,
            "pastro": null
        },
        "S230522a": {
            "pipelines": {
                "gstlal": "G651776"
            },
            "far": null,
            "pastro": null
        },
        "S230513ar": {
            "pipelines": {
                "gstlal": "G651774"
            },
            "far": null,
            "pastro": null
        }

    Returns
    =======
    dict :
        updated smap with fars and pastros computed from min FAR and min P
        terrestrial across pipelines, e.g.,

        "S230824d": {
            "pipelines": {
                "gstlal": "G651775"
            },
            "far": 1.611950746698324e-13,
            "pastro": null
        },
        "S230522a": {
            "pipelines": {
                "gstlal": "G651776"
            },
            "far": 3.171756306943489e-16,
            "pastro": {
                "BBH": 1.0,
                "BNS": 0.0,
                "NSBH": 0.0,
                "Terrestrial": 0.0
            }
        },
        "S230513ar": {
            "pipelines": {
                "gstlal": "G651774"
            },
            "far": 2.016192919429939e-07,
            "pastro": null
        }
    """
    smap = smap.copy()
    for se, d in smap.items():
        fars = []
        pastros = []
        for p, g in d["pipelines"].items():
            fars.append(gdb.event(g).json()["far"])
            pastro_files = [
                k for k, f in gdb.files(g).json().items() if k.endswith("p_astro.json")
            ]
            if len(pastro_files) > 0:
                assert len(pastro_files) == 1
                r = gdb.files(g, pastro_files[0])
                pastros.append(r.json())
        if len(fars) > 0:
            d["far"] = min(fars)
        if len(pastros) > 0:
            d["pastro"] = sorted(pastros, key=lambda d: d["Terrestrial"])[0]
    return smap


def smap_from_gracedb_query(
    gdb: GWTCGraceDB,
    query: str,
) -> Dict:
    """For each superevent satisfying query conditions,
    produce a map from s-name to g-event data for online pipeline preferred events

    Parameters
    ==========
    gdb : `gwtc.gwtc_gracedb.GWTCGraceDB`
        A rest API GraceDB instance modified for GWTC support
    query : string, required
        Gracedb query to extract super events

    Returns
    =======
    dict :
        A dictionary with s-events as keys,
        and values of dictionaries, e.g.,

        "S230824d": {
            "pipelines": {
                "gstlal": "G651775"
            },
            "far": 1.611950746698324e-13,
            "pastro": null
        },
        "S230522a": {
            "pipelines": {
                "gstlal": "G651776"
            },
            "far": 3.171756306943489e-16,
            "pastro": {
                "BBH": 1.0,
                "BNS": 0.0,
                "NSBH": 0.0,
                "Terrestrial": 0.0
            }
        },
        "S230513ar": {
            "pipelines": {
                "gstlal": "G651774"
            },
            "far": 2.016192919429939e-07,
            "pastro": null
        }
    """
    logging.info(f"Querying gracedb with query: {query}")
    smap = {}
    for se in gdb.superevents(query=query):
        smap[se["superevent_id"]] = {
            "pipelines": {
                pipeline: data["graceid"]
                for pipeline, data in se["pipeline_preferred_events"].items()
            },
            "far": None,
            "pastro": None,
        }
    logging.info("found the following event map:\n" + json.dumps(smap, indent=4))
    smap = update_far_and_pastro_in_smap(gdb, smap)
    return smap


def _diff(smap1, smap2):
    # calculate the differences and similarities in super event ids
    newsid = set(smap2) - set(smap1)
    delsid = set(smap1) - set(smap2)
    sameid = set(smap2) & set(smap1)
    changedsid = set()
    for sid in sameid:
        assert set(smap2[sid]) == set(smap1[sid])
        g1 = set(smap1[sid]["pipelines"].values()) | set((smap1[sid]["far"],))
        if smap1[sid]["pastro"]:
            g1 |= set(smap1[sid]["pastro"].items())
        g2 = set(smap2[sid]["pipelines"].values()) | set((smap2[sid]["far"],))
        if smap2[sid]["pastro"]:
            g2 |= set(smap2[sid]["pastro"].items())
        if g1 != g2:
            changedsid.add(sid)
    return newsid, delsid, sameid, changedsid


def catalog_delta(
    gdb: GWTCGraceDB, v1: int = None, v2: int = None, number: str = "4"
) -> Tuple[set, set, set, str]:
    """Highlight changes between v2 and v1 of catalog

    Parameters
    ==========
    gdb : `gwtc.gwtc_gracedb.GWTCGraceDB`
        A rest API GraceDB instance modified for GWTC support
    v1 : int
        The version number of the base catalog for comparison
    v2 : int
        The version number of the updated catalog for comparison
    number : str, optional
        The GWTC number ("4" for the time being)

    Returns
    =======
    set
        The set of newly added superevents sids
    set
        The set of removed superevent sids
    set
        The set of superevent sids which have had their contents updated
    str
        The human readable summary of the changes
    """
    if v1 is not None and v2 is not None:
        smap1 = gdb.gwtc_get(number, v1).json()["gwtc_superevents"]
        smap2 = gdb.gwtc_get(number, v2).json()["gwtc_superevents"]
    elif v1 is None and v2 is None:
        smap2 = gdb.gwtc_get(number, "latest").json()
        v2 = smap2["version"]
        smap2 = smap2["gwtc_superevents"]
        v1 = v2 - 1
        smap1 = gdb.gwtc_get(number, v1).json()["gwtc_superevents"]
    else:
        raise ValueError("v1 and v2 must both be integers or both be None")
    newsid, delsid, sameid, changedsid = _diff(smap1, smap2)

    outstr = """
- new s events in version %d: %s
- deleted s events in version %d: %s
- changed s events in version %d:
""" % (
        v2,
        sorted(newsid),
        v2,
        sorted(delsid),
        v2,
    )
    for sid in changedsid:
        outstr += f"\n\t{sid}:\n"
        for pipeline in set(smap1[sid]) | set(smap2[sid]):
            g1 = None if pipeline not in smap1[sid] else smap1[sid][pipeline]
            g2 = None if pipeline not in smap2[sid] else smap2[sid][pipeline]
            if g1 != g2:
                outstr += f"\t\t{pipeline}: {g1} -> {g2}\n"
    return newsid, delsid, changedsid, outstr


def gwtc_update_pipeline_events(
    gdb: GWTCGraceDB, pmap: dict, pipeline: str, number: str = "4", reset: bool = False
) -> requests.models.Response:
    """Update GWTC "number" for "pipeline" with provided changes

    Parameters
    ==========
    gdb : `gwtc.gwtc_gracedb.GWTCGraceDB`
        A rest API GraceDB instance modified for GWTC support
    pmap : dict
        The pipeline map to perform the update with, of the form, e.g.,

        "S230824d": {
            "pipelines": {
                "gstlal": "G651775"
            },
            "far": 1.611950746698324e-13,
            "pastro": null
        },
        "S230522a": {
            "pipelines": {
                "gstlal": "G651776"
            },
            "far": 3.171756306943489e-16,
            "pastro": {
                "BBH": 1.0,
                "BNS": 0.0,
                "NSBH": 0.0,
                "Terrestrial": 0.0
            }
        },
        "S230513ar": {
            "pipelines": {
                "gstlal": "G651774"
            },
            "far": 2.016192919429939e-07,
            "pastro": null
        }

    pipeline : str
        The pipeline for which to perform the update
    number : int, optional
        The GWTC number (4 for the time being)
    reset : bool
        If "reset" is provided, superevents
        that exist in the catalog but not in the pmap provided will have their gevents
        for "pipeline" removed. If a superevent has no pipelines left, it is deleted.

    Returns
    =======
    :class:`requests.models.Response`
    """
    # Sanity check that only pipeline events are provided and that they exist in
    # gracedb with the correct pipeline
    for s, d in pmap.items():
        assert set(d["pipelines"]) == set((pipeline,))
        gid = d["pipelines"][pipeline]
        assert gdb.event(gid).json()["pipeline"] == pipeline

    # get the current catalog
    try:
        smap = gdb.gwtc_get(number, "latest")
        smap = smap.json()["gwtc_superevents"]
    except requests.exceptions.HTTPError as e:
        logging.warn(
            "Couldn't retrieve previous version of catalog. Setting to empty dictionary. %s"
            % e
        )
        smap = {}

    new_sids, delete_sids, _, update_sids = _diff(smap, pmap)

    # add new super events that were not there previously
    for sid in new_sids:
        smap[sid] = pmap[sid]
    # update sids that are in both
    for sid in update_sids:
        smap[sid]["pipelines"].update(pmap[sid]["pipelines"])
    if reset:
        # delete pipeline gevents from superevents not in the provided map
        for sid in delete_sids:
            if pipeline in smap[sid]["pipelines"]:
                del smap[sid]["pipelines"][pipeline]
            # delete super events that no longer have any pipelines
            if len(smap[sid]["pipelines"]) == 0:
                del smap[sid]

    update_far_and_pastro_in_smap(gdb, smap)
    return gdb.gwtc_create(smap, number=number)
